{-# OPTIONS -#include <time.h> #-}
{-# OPTIONS -#include <libpq-fe.h> #-}
{-# OPTIONS -#include <postgres.h> #-}
{-# OPTIONS -#include <catalog/pg_type.h> #-}
{-# LINE 1 "PostgreSQL.hsc" #-}
-----------------------------------------------------------------------------------------
{-# LINE 2 "PostgreSQL.hsc" #-}
{-| Module      :  Database.HSQL.PostgreSQL
    Copyright   :  (c) Krasimir Angelov 2003
    License     :  BSD-style

    Maintainer  :  ka2_mail@yahoo.com
    Stability   :  provisional
    Portability :  portable

    The module provides interface to PostgreSQL database
-}
-----------------------------------------------------------------------------------------

module Database.HSQL.PostgreSQL(connect, module Database.HSQL) where

import Database.HSQL
import Database.HSQL.Types
import Data.Dynamic
import Data.Char
import Foreign
import Foreign.C
import Control.Exception (throwDyn, catchDyn, dynExceptions, Exception(..))
import Control.Monad(when,unless,mplus)
import Control.Concurrent.MVar
import System.Time
import System.IO.Unsafe
import Text.ParserCombinators.ReadP
import Text.Read
import Numeric


{-# LINE 32 "PostgreSQL.hsc" #-}

{-# LINE 33 "PostgreSQL.hsc" #-}

{-# LINE 34 "PostgreSQL.hsc" #-}

{-# LINE 35 "PostgreSQL.hsc" #-}

type PGconn = Ptr ()
type PGresult = Ptr ()
type ConnStatusType = Word32
{-# LINE 39 "PostgreSQL.hsc" #-}
type ExecStatusType = Word32
{-# LINE 40 "PostgreSQL.hsc" #-}
type Oid = Word32
{-# LINE 41 "PostgreSQL.hsc" #-}

foreign import ccall "libpq-fe.h PQsetdbLogin" pqSetdbLogin :: CString -> CString -> CString -> CString -> CString -> CString -> CString -> IO PGconn
foreign import ccall "libpq-fe.h PQstatus" pqStatus :: PGconn -> IO ConnStatusType
foreign import ccall "libpq-fe.h PQerrorMessage"  pqErrorMessage :: PGconn -> IO CString
foreign import ccall "libpq-fe.h PQfinish" pqFinish :: PGconn -> IO ()
foreign import ccall "libpq-fe.h PQexec" pqExec :: PGconn -> CString -> IO PGresult
foreign import ccall "libpq-fe.h PQresultStatus" pqResultStatus :: PGresult -> IO ExecStatusType
foreign import ccall "libpq-fe.h PQresStatus" pqResStatus :: ExecStatusType -> IO CString
foreign import ccall "libpq-fe.h PQresultErrorMessage" pqResultErrorMessage :: PGresult -> IO CString
foreign import ccall "libpq-fe.h PQnfields" pgNFields :: PGresult -> IO Int
foreign import ccall "libpq-fe.h PQntuples" pqNTuples :: PGresult -> IO Int
foreign import ccall "libpq-fe.h PQfname" pgFName :: PGresult -> Int -> IO CString
foreign import ccall "libpq-fe.h PQftype" pqFType :: PGresult -> Int -> IO Oid
foreign import ccall "libpq-fe.h PQfmod" pqFMod :: PGresult -> Int -> IO Int
foreign import ccall "libpq-fe.h PQfnumber" pqFNumber :: PGresult -> CString -> IO Int
foreign import ccall "libpq-fe.h PQgetvalue" pqGetvalue :: PGresult -> Int -> Int -> IO CString
foreign import ccall "libpq-fe.h PQgetisnull" pqGetisnull :: PGresult -> Int -> Int -> IO Int

foreign import ccall "strlen" strlen :: CString -> IO Int

-----------------------------------------------------------------------------------------
-- Connect/Disconnect
-----------------------------------------------------------------------------------------

-- | Makes a new connection to the database server.
connect :: String   -- ^ Server name
        -> String   -- ^ Database name
        -> String   -- ^ User identifier
        -> String   -- ^ Authentication string (password)
        -> IO Connection
connect server database user authentication = do
	pServer <- newCString server
	pDatabase <- newCString database
	pUser <- newCString user
	pAuthentication <- newCString authentication
	pConn <- pqSetdbLogin pServer nullPtr nullPtr nullPtr pDatabase pUser pAuthentication
	free pServer
	free pUser
	free pAuthentication
	status <- pqStatus pConn
	unless (status == (0))  (do
{-# LINE 82 "PostgreSQL.hsc" #-}
		errMsg <- pqErrorMessage pConn >>= peekCString
		pqFinish pConn
		throwDyn (SqlError {seState="C", seNativeError=fromIntegral status, seErrorMsg=errMsg}))
	refFalse <- newMVar False
	let connection = Connection
		{ connDisconnect = pqFinish pConn
		, connExecute = execute pConn
		, connQuery = query connection pConn
		, connTables = tables connection pConn
		, connDescribe = describe connection pConn
		, connBeginTransaction = execute pConn "begin"
		, connCommitTransaction = execute pConn "commit"
		, connRollbackTransaction = execute pConn "rollback"
		, connClosed = refFalse
		}
	return connection
	where
		execute :: PGconn -> String -> IO ()
		execute pConn sqlExpr = do
			pRes <- withCString sqlExpr (pqExec pConn)
			when (pRes==nullPtr) (do
				errMsg <- pqErrorMessage pConn >>= peekCString
				throwDyn (SqlError {seState="E", seNativeError=(7), seErrorMsg=errMsg}))
{-# LINE 105 "PostgreSQL.hsc" #-}
			status <- pqResultStatus pRes
			unless (status == (1) || status == (2)) (do
{-# LINE 107 "PostgreSQL.hsc" #-}
				errMsg <- pqResultErrorMessage pRes >>= peekCString
				throwDyn (SqlError {seState="E", seNativeError=fromIntegral status, seErrorMsg=errMsg}))
			return ()

		query :: Connection -> PGconn -> String -> IO Statement
		query conn pConn query = do
			pRes <- withCString query (pqExec pConn)
			when (pRes==nullPtr) (do
				errMsg <- pqErrorMessage pConn >>= peekCString
				throwDyn (SqlError {seState="E", seNativeError=(7), seErrorMsg=errMsg}))
{-# LINE 117 "PostgreSQL.hsc" #-}
			status <- pqResultStatus pRes
			unless (status == (1) || status == (2)) (do
{-# LINE 119 "PostgreSQL.hsc" #-}
				errMsg <- pqResultErrorMessage pRes >>= peekCString
				throwDyn (SqlError {seState="E", seNativeError=fromIntegral status, seErrorMsg=errMsg}))
			defs <- if status ==  (2) then pgNFields pRes >>= getFieldDefs pRes 0 else return []
{-# LINE 122 "PostgreSQL.hsc" #-}
			countTuples <- pqNTuples pRes;
			tupleIndex <- newMVar (-1)
			refFalse <- newMVar False
			return (Statement
			              { stmtConn   = conn
			              , stmtClose  = return ()
			              , stmtFetch  = fetch tupleIndex countTuples
			              , stmtGetCol = getColValue pRes tupleIndex countTuples
			              , stmtFields = defs
			              , stmtClosed = refFalse
			              })
			where
				getFieldDefs pRes i n
					| i >= n = return []
					| otherwise = do
						name <- pgFName pRes i	>>= peekCString
						dataType <- pqFType pRes i
						modifier <- pqFMod pRes i
						defs <- getFieldDefs pRes (i+1) n
						return ((name,mkSqlType dataType modifier,True):defs)

		mkSqlType :: Oid -> Int -> SqlType
		mkSqlType (1042)    size = SqlChar (size-4)
{-# LINE 145 "PostgreSQL.hsc" #-}
		mkSqlType (1043)   size = SqlVarChar (size-4)
{-# LINE 146 "PostgreSQL.hsc" #-}
		mkSqlType (19)      size = SqlVarChar 31
{-# LINE 147 "PostgreSQL.hsc" #-}
		mkSqlType (25)      size = SqlText
{-# LINE 148 "PostgreSQL.hsc" #-}
		mkSqlType (1700)   size = SqlNumeric ((size-4) `div` 0x10000) ((size-4) `mod` 0x10000)
{-# LINE 149 "PostgreSQL.hsc" #-}
		mkSqlType (21)      size = SqlSmallInt
{-# LINE 150 "PostgreSQL.hsc" #-}
		mkSqlType (23)      size = SqlInteger
{-# LINE 151 "PostgreSQL.hsc" #-}
		mkSqlType (700)    size = SqlReal
{-# LINE 152 "PostgreSQL.hsc" #-}
		mkSqlType (701)    size = SqlDouble
{-# LINE 153 "PostgreSQL.hsc" #-}
		mkSqlType (16)      size = SqlBit
{-# LINE 154 "PostgreSQL.hsc" #-}
		mkSqlType (1560)       size = SqlBinary size
{-# LINE 155 "PostgreSQL.hsc" #-}
		mkSqlType (1562)    size = SqlVarBinary size
{-# LINE 156 "PostgreSQL.hsc" #-}
		mkSqlType (17)     size = SqlTinyInt
{-# LINE 157 "PostgreSQL.hsc" #-}
		mkSqlType (20)      size = SqlBigInt
{-# LINE 158 "PostgreSQL.hsc" #-}
		mkSqlType (1082)      size = SqlDate
{-# LINE 159 "PostgreSQL.hsc" #-}
		mkSqlType (1083)      size = SqlTime
{-# LINE 160 "PostgreSQL.hsc" #-}
		mkSqlType (1266)    size = SqlTimeTZ
{-# LINE 161 "PostgreSQL.hsc" #-}
		mkSqlType (702)   size = SqlAbsTime
{-# LINE 162 "PostgreSQL.hsc" #-}
		mkSqlType (703)   size = SqlRelTime
{-# LINE 163 "PostgreSQL.hsc" #-}
		mkSqlType (1186)  size = SqlTimeInterval
{-# LINE 164 "PostgreSQL.hsc" #-}
		mkSqlType (704) size = SqlAbsTimeInterval
{-# LINE 165 "PostgreSQL.hsc" #-}
		mkSqlType (1114)	size = SqlDateTime
{-# LINE 166 "PostgreSQL.hsc" #-}
		mkSqlType (1184)	size = SqlDateTimeTZ
{-# LINE 167 "PostgreSQL.hsc" #-}
		mkSqlType (790)      size = SqlMoney
{-# LINE 168 "PostgreSQL.hsc" #-}
		mkSqlType (869)      size = SqlINetAddr
{-# LINE 169 "PostgreSQL.hsc" #-}
		mkSqlType (829)          size = SqlMacAddr		-- hack
{-# LINE 170 "PostgreSQL.hsc" #-}
		mkSqlType (650)      size = SqlCIDRAddr
{-# LINE 171 "PostgreSQL.hsc" #-}
		mkSqlType (600)     size = SqlPoint
{-# LINE 172 "PostgreSQL.hsc" #-}
		mkSqlType (601)      size = SqlLSeg
{-# LINE 173 "PostgreSQL.hsc" #-}
		mkSqlType (602)      size = SqlPath
{-# LINE 174 "PostgreSQL.hsc" #-}
		mkSqlType (603)       size = SqlBox
{-# LINE 175 "PostgreSQL.hsc" #-}
		mkSqlType (604)   size = SqlPolygon
{-# LINE 176 "PostgreSQL.hsc" #-}
		mkSqlType (628)      size = SqlLine
{-# LINE 177 "PostgreSQL.hsc" #-}
		mkSqlType (718)    size = SqlCircle
{-# LINE 178 "PostgreSQL.hsc" #-}
		mkSqlType tp   size = SqlUnknown (fromIntegral tp)

		getFieldValue stmt colNumber fieldDef v = do
			mb_v <- stmtGetCol stmt colNumber fieldDef fromNonNullSqlCStringLen
			return (case mb_v of { Nothing -> v; Just a -> a })

		tables :: Connection -> PGconn -> IO [String]
		tables connection pConn = do
			stmt <- query connection pConn "select relname from pg_class where relkind='r' and relname !~ '^pg_'"
			collectRows (\s -> getFieldValue  s 0 ("relname", SqlVarChar 0, False) "") stmt

		describe :: Connection -> PGconn -> String -> IO [FieldDef]
		describe connection pConn table = do
			stmt <- query connection pConn
					("select attname, atttypid, atttypmod, attnotnull " ++
					 "from pg_attribute as cols join pg_class as ts on cols.attrelid=ts.oid " ++
					 "where cols.attnum > 0 and ts.relname='"++table++"'")
			collectRows getColumnInfo stmt
			where
				getColumnInfo stmt = do
					column_name <- getFieldValue stmt 0 ("attname", SqlVarChar 0, False) ""
					(data_type::Int) <- getFieldValue stmt 1 ("atttypid", SqlInteger, False) 0
					(type_mod::Int) <- getFieldValue stmt 2 ("atttypmod", SqlInteger, False) 0
					(notnull::Bool) <- getFieldValue stmt 3 ("attnotnull", SqlBit, False) False
					let sqlType = mkSqlType (fromIntegral data_type) (fromIntegral type_mod)
					return (column_name, sqlType, not notnull)

		fetch :: MVar Int -> Int -> IO Bool
		fetch tupleIndex countTuples =
			modifyMVar tupleIndex (\index -> return (index+1,index < countTuples-1))

		getColValue :: PGresult -> MVar Int -> Int -> Int -> FieldDef -> (SqlType -> CString -> Int -> IO (Maybe a)) -> IO (Maybe a)
		getColValue pRes tupleIndex countTuples colNumber  (name,sqlType,nullable) f = do
			index <- readMVar tupleIndex
			when (index >= countTuples) (throwDyn SqlNoData)
			isnull <- pqGetisnull pRes index colNumber
			if isnull == 1
				then return Nothing
				else do
					pStr <- pqGetvalue pRes index colNumber
					strLen <- strlen pStr
					mb_value <- f sqlType pStr strLen
					case mb_value of
						Just v   -> return (Just v)
						Nothing -> throwDyn (SqlBadTypeCast name sqlType)
